// ------------------------------------------------------------
// Machinery project - math library
// ------------------------------------------------------------
// (c) 2000 by Sascha 'EvilOne' Salevsky
// E-mail: Sascha.Salevsky@Ibykus.de
// ------------------------------------------------------------

#ifndef _mathematics_h_
#error "Dont include Quaternion directly!!! Use Mathematics.h instead."
#endif

// Constructors

inline Quaternion::Quaternion()
{}

inline Quaternion::Quaternion(const Quaternion& other)
{
	x = other.x; y = other.y; z = other.z; w = other.w;
}

inline Quaternion::Quaternion(const float* other)
{
	x = other[0]; y = other[1]; z = other[2]; w = other[3];
}

inline Quaternion::Quaternion(float _x, float _y, float _z, float _w)
{
	x = _x; y = _y; z = _z; w = _w;
}

// Assignment

inline Quaternion& Quaternion::operator=(const Quaternion& other)
{
	x = other.x; y = other.y; z = other.z; w = other.w;
	return *this;
}

inline void Quaternion::Assign(const Quaternion& other)
{
	x = other.x; y = other.y; z = other.z; w = other.w;
}

inline void Quaternion::Assign(const float* other)
{
	x = other[0]; y = other[1]; z = other[2]; w = other[3];
}

inline void Quaternion::Assign(float _x, float _y, float _z, float _w)
{
	x = _x; y = _y; z = _z; w = _w;
}

// Simple math

inline void Quaternion::Invert()
{
	x = -x; y = -y; z = -z;
}

inline void Quaternion::Invert(const Quaternion& other)
{
	x = -other.x; y = -other.y; z = -other.z;
	w = other.w;
}

inline Quaternion& Quaternion::operator-()
{
	x = -x; y = -y; z = -z;
	return *this;
}

inline float Quaternion::Length() const
{
	return (float)sqrt(w*w + x*x + y*y + z*z);
}

inline float Quaternion::LengthSqr() const
{
	return w*w + x*x + y*y + z*z;
}

inline void Quaternion::Normalize()
{
	float f = 1.0f / (w*w + x*x + y*y + z*z);
	x *= f; y *= f; z *= f; w *= f;
}

inline void Quaternion::Normalize(const Quaternion& other)
{
	float f = 1.0f / other.Length();
	x = other.x * f; y = other.y * f; z = other.z * f; w = other.w * f;
}

inline Quaternion& Quaternion::Mul(const Quaternion& other)
{
	Mul(*this, other);
	return *this;
}

inline Quaternion& Quaternion::Mul(const Quaternion& q1, const Quaternion& q2)
{
	const Vector3 v1 = &q1.x;
	const Vector3 v2 = &q2.x;

    float scalar = q1.w * q2.w - v1.Dot(v2);

    Vector3 va; va.Cross(v1, v2);
	Vector3 vb; vb.Mul(v1, q2.w);
	Vector3 vc; vc.Mul(v2, q1.w);
	va.Add(vb); va.Add(vc);

	x = va.x; y = va.y; z = va.z; w = scalar;

    Normalize();
	return *this;
}

inline Quaternion Quaternion::operator*(const Quaternion& other) const
{
	Quaternion tmp;
	tmp.Mul(*this, other);
	return tmp;
}

inline Quaternion& Quaternion::operator*=(const Quaternion& other)
{
	Mul(*this, other);
	return *this;
}

// Rotation

inline Quaternion& Quaternion::SetRotate(float angle, const Vector3& v)
{
	return SetRotate(angle, v.x, v.y, v.z);
}

inline Quaternion& Quaternion::Rotate(float angle, const Vector3& v)
{
	return Rotate(angle, v.x, v.y, v.z);
}

inline Quaternion& Quaternion::SetRotate(const Vector3& v)
{
	return SetRotate(v.x, v.y, v.z);
}

inline Quaternion& Quaternion::Rotate(const Vector3& v)
{
	return Rotate(v.x, v.y, v.z);
}

inline Quaternion& Quaternion::SetRotate(float angle, float _x, float _y, float _z)
{
    float s = SinD(angle / 2.0f);
    float c = CosD(angle / 2.0f);

    x = _x * s;
    y = _y * s;
    z = _z * s;
    w = c;
	Normalize();
	return *this;
}

inline Quaternion& Quaternion::Rotate(float angle, float _x, float _y, float _z)
{
	Quaternion tmp;
	tmp.SetRotate(angle, _x, _y, _z);
	Mul(*this, tmp);
	return *this;
}

inline Quaternion& Quaternion::SetRotate(float rx, float ry, float rz)
{
	Quaternion qy, qz;
	SetRotate(rx, 1.0f, 0.0f, 0.0f);
	qy.SetRotate(ry, 0.0f, 1.0f, 0.0f);
	qz.SetRotate(rz, 0.0f, 0.0f, 1.0f);
	Mul(*this, qy);
	Mul(*this, qz);
	return *this;
}

inline Quaternion& Quaternion::Rotate(float rx, float ry, float rz)
{
	Quaternion tmp;
	tmp.SetRotate(rx, ry, rz);
	Mul(*this, tmp);
	return *this;
}
